"""
Molecule
========

#. :class:`.Molecule`

Molecule class.

"""

from rdkit.Chem import AllChem as Chem
from rdkit.Chem.Descriptors3D import NPR1, NPR2
from rdkit.Geometry import rdGeometry

import numpy as np

from .ellipsefitter import EllipsoidTool
from .utilities import plot_ellipsoid


class Molecule:
    """
    Molecule to calculate size of.

    """

    def __init__(self, rdkitmol, conformers):
        """
        Initialize a :class:`Molecule` instance.

        Parameters
        ----------
        rdkitmol : :class:`RDKit.Molecule`
            RDKit molecule to get size of.

        conformers : :class:`iterable`
            Iterable of the conformer ids used to access the conformers
            in the rdkit molecule.

        """

        self._rdkitmol = rdkitmol
        self._conformers = conformers

    def get_inertial_ratios(self):
        """
        Get inertial 3D descriptors for all conformers in mol.

        Returns
        -------
        conf_ratios : :class:`dict` of :class:`tuple`
            Dictionary of ratio_1 and ratio_2 of all conformers.
            Key is conformer id. Ratio 1 is I1/I3, ratio 2 is I2/I3.

        """

        conf_ratios = {}
        for cid in self._conformers:
            conf_ratios[cid] = (
                NPR1(self._rdkitmol, confId=cid),
                NPR2(self._rdkitmol, confId=cid),
            )

        return conf_ratios

    def get_molecule_shape(
        self,
        conformer,
        cid,
        vdwscale,
        boxmargin,
        spacing
    ):
        """
        Get the shape of a conformer of a molecule as a grid.

        """
        box = Chem.ComputeConfBox(conformer)
        sideLen = (
            box[1].x-box[0].x + 2*boxmargin,
            box[1].y-box[0].y + 2*boxmargin,
            box[1].z-box[0].z + 2*boxmargin,
        )
        shape = rdGeometry.UniformGrid3D(
            2*sideLen[0],
            2*sideLen[1],
            2*sideLen[2],
            spacing=spacing
        )
        Chem.EncodeShape(
            self._rdkitmol,
            shape,
            confId=cid,
            ignoreHs=False,
            vdwScale=vdwscale
        )
        return box, sideLen, shape

    def get_hitpoints(self, shape):
        """
        Get points with value > 2 that are within vdw shape.

        """

        hit_points = []
        for idx in range(shape.GetSize()):
            value = shape.GetVal(idx)
            if value > 2:
                pt = shape.GetGridPointLoc(idx)
                point = np.array([pt.x, pt.y, pt.z])
                hit_points.append(point)
        hit_points = np.asarray(hit_points)

        return hit_points

    def get_ellipsoids(
        self,
        vdwscale,
        boxmargin,
        spacing,
    ):
        """
        Get min volume ellipsoids for all conformers in mol.

        Good values:
            vdwScale:0.9
            boxMargin:4.0
            spacing:0.5

        Returns
        -------
        conf_ellipsoids : :class:`dict` of :class:`tuple`
            Dictionary of conformer ellipsoids, key is conformer id.
            Value is (center, diameters, rotation matrix).

        """

        conf_ellipsoids = {}
        for cid in self._conformers:
            conformer = self._rdkitmol.GetConformer(cid)
            box, sideLen, shape = self.get_molecule_shape(
                conformer=conformer,
                cid=cid,
                vdwscale=vdwscale,
                boxmargin=boxmargin,
                spacing=spacing,
            )

            hit_points = self.get_hitpoints(shape)

            # Find the ellipsoid that envelopes all hit points.
            ET = EllipsoidTool()
            (center, radii, rotation) = ET.get_min_vol_ellipse(
                points=hit_points,
                tolerance=0.01,
            )
            diameters = list(np.sort(radii)*2)

            conf_ellipsoids[cid] = (center, diameters, rotation)

        return conf_ellipsoids

    def draw_conformer_hitpoints(
        self,
        cid,
        center,
        diameter,
        rotation,
        vdwscale,
        boxmargin,
        spacing,
        filename,
    ):
        """
        Draw hitpoints of ellipsoid fit in plot.

        """

        conformer = self._rdkitmol.GetConformer(cid)
        box, sideLen, shape = self.get_molecule_shape(
            conformer=conformer,
            cid=cid,
            vdwscale=vdwscale,
            boxmargin=boxmargin,
            spacing=spacing,
        )

        hit_points = self.get_hitpoints(shape)
        fig, ax = plot_ellipsoid(
            center,
            diameter,
            rotation,
        )
        ax.scatter(
            hit_points[:, 0], hit_points[:, 1], hit_points[:, 2],
            color='g',
            marker='x',
            edgecolor=None,
            s=50,
            alpha=0.5,
        )

        fig.tight_layout()
        fig.savefig(filename, dpi=720, bbox_inches='tight')

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} at {id(self)}> '
            f'with {len(self._conformers)} conformers'
        )



"""
EllisoidTool
============

From: https://github.com/minillinim/ellipsoid

ellipsoid.py
Copyright (C) 2018  Michael Imelfort

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

#. :class:`.EllipsoidTool`

EllipsoidTool class.

"""


from __future__ import division
import numpy as np


class EllipsoidTool:
    """
    EllipsoidTool for fitting ellipsoids.

    """

    def __init__(self):
        pass

    def get_min_vol_ellipse(
        self,
        points,
        tolerance=0.01,
    ):
        """
        Find the minimum volume ellipsoid which holds all the points.

        Based on work by Nima Moshtagh
        http://www.mathworks.com/matlabcentral/fileexchange/9542
        and also by looking at:
        http://cctbx.sourceforge.net/current/python/
        scitbx.math.minimum_covering_ellipsoid.html
        Which is based on the first reference anyway!

        Parameters
        ----------
        points : :class:`np.ndarray`
            (N, 3) numpy array of N points.

        tolerance : :class:`float`, optional
            Tolerance used in Khachiyan algorithm.

        Returns
        ------
        :class:`tuple` of :class:`np.ndarray`
            Tuple of arrays containing coordinates of ellipsoid center,
            ellipsoid radii, ellipsoid rotation matrix.

        """

        (N, d) = np.shape(points)
        d = float(d)

        # Q will be our working array
        Q = np.vstack([np.copy(points.T), np.ones(N)])
        QT = Q.T

        # initializations
        err = 1.0 + tolerance
        u = (1.0 / N) * np.ones(N)

        # Khachiyan Algorithm
        count = 0
        while err > tolerance:
            V = np.dot(Q, np.dot(np.diag(u), QT))
            # M the diagonal vector of an NxN matrix
            M = np.diag(np.dot(QT, np.dot(np.linalg.inv(V), Q)))
            j = np.argmax(M)
            maximum = M[j]
            step_size = (maximum - d - 1.0)
            step_size = step_size / ((d + 1.0) * (maximum - 1.0))
            new_u = (1.0 - step_size) * u
            new_u[j] += step_size
            err = np.linalg.norm(new_u - u)
            u = new_u
            count += 1

        # center of the ellipse
        center = np.dot(points.T, u)

        # the A matrix for the ellipse
        A = np.linalg.inv(
            np.dot(points.T, np.dot(np.diag(u), points)) -
            np.array([[a * b for b in center] for a in center])
        ) / d

        # Get the values we'd like to return
        U, s, rotation = np.linalg.svd(A)
        radii = 1.0/np.sqrt(s)

        return (center, radii, rotation)



import pytest
import numpy as np
import molellipsize as mes


def test_get_min_vol_ellipse(points):
    ET = mes.EllipsoidTool()

    test_center, test_radii, test_rotation = ET.get_min_vol_ellipse(
        points.points
    )

    print(points.center, points.radii, points.rotation)
    print(test_center, test_radii, test_rotation)

    for known, test_cent_dim in zip(points.center, test_center):
        assert np.isclose(known, test_cent_dim, rtol=0, atol=1E-1)

    for known, test_radius in zip(points.radii, test_radii):
        assert np.isclose(known, test_radius, rtol=0, atol=1E-2)

    assert np.allclose(
        points.rotation, test_rotation, rtol=0, atol=1E-3
    )



import pytest
import numpy as np
import molellipsize as mes


def test_get_inertial_ratios(molecule):
    mes_mol = mes.Molecule(molecule.molecule, molecule.conformers)
    conf_ratios = mes_mol.get_inertial_ratios()
    print(conf_ratios)
    test_ratio_1, test_ratio_2 = conf_ratios[0]
    assert np.isclose(
        test_ratio_1, molecule.inertial_ratios[0], rtol=0, atol=1E-1
    )
    assert np.isclose(
        test_ratio_2, molecule.inertial_ratios[1], rtol=0, atol=1E-1
    )


def test_get_ellipsoids(molecule):
    mes_mol = mes.Molecule(molecule.molecule, molecule.conformers)
    conf_ellipsoids = mes_mol.get_ellipsoids(1.0, 4.0, 0.5)

    test_center = conf_ellipsoids[0][0]
    for known, test_cent_dim in zip(
        molecule.ellipse_center,
        test_center
    ):
        assert np.isclose(known, test_cent_dim, rtol=0, atol=1E-1)

    test_diameters = conf_ellipsoids[0][1]
    print(test_diameters)
    for known, test_diam in zip(
        molecule.ellipse_diameters,
        test_diameters
    ):
        assert np.isclose(known, test_diam, rtol=0, atol=1E-5)



import pytest
from rdkit.Chem import AllChem as Chem
import molellipsize as mes


class CaseData:
    """
    A test case.

    Attributes
    ----------
    atom_atomic_nums : :class:`list`
        List of atom atomic numbers.

    pos_mat : :class:`iter` of :class:`list`
        List of atom coordinates.

    ellipse_center : :class:`tuple`
        Center of test molecule's ellipsoid.

    ellipse_diameters : :class:`tuple`
        Diameters of test molecule's ellipsoid.

    inertial_ratios : :class:`tuple`
        Inertial ratios of the test molecule.

    """

    def __init__(
        self,
        atom_atomic_nums,
        pos_mat,
        ellipse_center,
        ellipse_diameters,
        inertial_ratios
    ):

        # Make rdkit molecule.
        mol = Chem.EditableMol(Chem.Mol())
        for an in atom_atomic_nums:
            rdkit_atom = Chem.Atom(an)
            mol.AddAtom(rdkit_atom)

        mol = mol.GetMol()
        rdkit_conf = Chem.Conformer(len(atom_atomic_nums))
        for atom_id, atom_coord in enumerate(pos_mat):
            rdkit_conf.SetAtomPosition(atom_id, atom_coord)
            mol.GetAtomWithIdx(atom_id).SetNoImplicit(True)
        mol.AddConformer(rdkit_conf)

        self.molecule = mol
        self.conformers = [0]
        self.ellipse_center = ellipse_center
        self.ellipse_diameters = ellipse_diameters
        self.inertial_ratios = inertial_ratios


@pytest.fixture(
    params=(
        # Oblate, in xy dir
        CaseData(
            atom_atomic_nums=[1, 1, 1, 1],
            pos_mat=[
                [0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0],
            ],
            ellipse_center=(0.50, 0.50, 0.0),
            ellipse_diameters=(2.50904, 3.56057, 3.58628),
            inertial_ratios=(0.5, 0.5),
        ),
        # Oblate, but checks that the directionality plays no role,
        # by being in yz dir.
        CaseData(
            atom_atomic_nums=[1, 1, 1, 1],
            pos_mat=[
                [0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 1, 1],
            ],
            ellipse_center=(0.0, 0.50, 0.50),
            ellipse_diameters=(2.50904, 3.56057, 3.58628),
            inertial_ratios=(0.5, 0.5),
        ),
        # Sphere.
        CaseData(
            atom_atomic_nums=[1, 1, 1, 1, 1, 1, 1, 1],
            pos_mat=[
                [1, 1, 1], [1, -1, 1], [-1, 1, 1], [-1, -1, 1],
                [1, 1, -1], [1, -1, -1], [-1, 1, -1], [-1, -1, -1],
            ],
            ellipse_center=(0.0, 0.0, 0.0),
            ellipse_diameters=(5.29935, 5.33006, 5.35154),
            inertial_ratios=(1.0, 1.0),
        ),
        # Prolate.
        CaseData(
            atom_atomic_nums=[1, 1, 1, 1],
            pos_mat=[
                [5, 1, 0], [5, -1, 0], [-5, 1, 0], [-5, -1, 0],
            ],
            ellipse_center=(0.0, 0.0, 0.0),
            ellipse_diameters=(3.35897, 5.18066, 17.32586),
            inertial_ratios=(0.0, 1.0),
        ),
    ),
)
def molecule(request):
    return request.param



"""
This module defines general-purpose objects, functions and classes.

"""

import numpy as np
from rdkit.Chem import AllChem as Chem
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


def ETKDG_UFF_conformers(rdkitmol, num_conformers, randomseed):
    """
    Generate conformers with RDKit.

    """

    # Use a set randomSeed so that running the code multiple times
    # gives the same series of conformers.
    conformers = Chem.EmbedMultipleConfs(
        mol=rdkitmol,
        numConfs=num_conformers,
        useExpTorsionAnglePrefs=True,
        useBasicKnowledge=True,
        randomSeed=randomseed
    )

    for cid in conformers:
        Chem.UFFOptimizeMolecule(rdkitmol, confId=cid)

    return rdkitmol, conformers


def plot_shapes(ratios, filename):
    """
    Plot molecule intertial ratios.

    Saves figure at `filename`.

    """

    fig, ax = plt.subplots(figsize=(5, 5))

    ax.scatter(
        [ratios[i][0] for i in ratios],
        [ratios[i][1] for i in ratios],
        c='gold',
        edgecolors='white',
        alpha=1.0,
        s=80
    )

    ax.plot([0, 0.5, 1, 0], [1, 0.5, 1, 1], c='k', lw=2)
    ax.text(0.75, 1.03, 'sphere', fontsize=20)
    ax.text(0.4, 0.45, 'oblate', fontsize=20)
    ax.text(-0.05, 1.03, 'prolate', fontsize=20)

    # Set number of ticks for x-axis
    ax.tick_params(axis='both', which='major', labelsize=16)
    ax.set_xlabel('$I_1$ / $I_3$', fontsize=16)
    ax.set_ylabel('$I_2$ / $I_3$', fontsize=16)
    ax.set_xlim(-0.1, 1.1)
    ax.set_ylim(0.4, 1.1)

    fig.tight_layout()
    fig.savefig(filename, dpi=720, bbox_inches='tight')


def plot_diameters(diameters, filename):
    """
    Plot molecule diameters.

    Saves figure at `filename`.

    """

    d1 = [diameters[i][0] for i in diameters]
    d2 = [diameters[i][1] for i in diameters]
    d3 = [diameters[i][2] for i in diameters]

    fig, ax = plt.subplots(figsize=(8, 5))
    width = 0.2
    xlim = (0, round(max(d3)+1, 0))
    X_bins = np.arange(xlim[0], xlim[1], width)
    hist, bin_edges = np.histogram(a=d1, bins=X_bins)
    ax.bar(
        bin_edges[:-1],
        hist,
        align='edge',
        alpha=1.0,
        width=width,
        color='none',
        edgecolor='gold',
        linewidth=3,
        label='d1',
    )
    hist, bin_edges = np.histogram(a=d2, bins=X_bins)
    ax.bar(
        bin_edges[:-1],
        hist,
        align='edge',
        alpha=1.0,
        width=width,
        color='none',
        edgecolor='palegreen',
        linewidth=3,
        label='d2',
    )
    hist, bin_edges = np.histogram(a=d3, bins=X_bins)
    ax.bar(
        bin_edges[:-1],
        hist,
        align='edge',
        alpha=1.0,
        width=width,
        color='none',
        edgecolor='coral',
        linewidth=3,
        label='d3',
    )
    # Set number of ticks for x-axis
    ax.tick_params(axis='both', which='major', labelsize=16)
    ax.set_xlabel(r'molecular diameter [$\mathrm{\AA}$]', fontsize=16)
    ax.set_ylabel('count', fontsize=16)
    ax.set_xlim(xlim)
    ax.set_ylim(0, None)
    ax.legend(fontsize=16)

    fig.tight_layout()
    fig.savefig(filename, dpi=720, bbox_inches='tight')


def plot_ellipsoid(
    center,
    diameter,
    rotation,
    plotAxes=False,
    cageColor='k',
    cageAlpha=0.2,
):
    """
    Plot an ellipsoid.

    Returns
    -------
    fig : class:`matplotlib.figure`
        Figure object with ellipsoid.

    ax : class:`matplotlib.axis`
        Axis object with ellipsoid.

    """

    radii = [i/2 for i in diameter]

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    u = np.linspace(0.0, 2.0 * np.pi, 100)
    v = np.linspace(0.0, np.pi, 100)

    # cartesian coordinates that correspond to the spherical
    # angles:
    x = radii[0] * np.outer(np.cos(u), np.sin(v))
    y = radii[1] * np.outer(np.sin(u), np.sin(v))
    z = radii[2] * np.outer(np.ones_like(u), np.cos(v))
    # rotate accordingly
    for i in range(len(x)):
        for j in range(len(x)):
            [x[i, j], y[i, j], z[i, j]] = np.dot(
                [x[i, j], y[i, j], z[i, j]],
                rotation
            ) + center

    if plotAxes:
        # make some purdy axes
        axes = np.array([
            [radii[0], 0.0, 0.0],
            [0.0, radii[1], 0.0],
            [0.0, 0.0, radii[2]]
        ])
        # rotate accordingly
        for i in range(len(axes)):
            axes[i] = np.dot(axes[i], rotation)

        # plot axes
        for p in axes:
            X3 = np.linspace(-p[0], p[0], 100) + center[0]
            Y3 = np.linspace(-p[1], p[1], 100) + center[1]
            Z3 = np.linspace(-p[2], p[2], 100) + center[2]
            ax.plot(X3, Y3, Z3, color=cageColor)

    # plot ellipsoid
    ax.plot_wireframe(
        x, y, z,
        rstride=4,
        cstride=4,
        color=cageColor,
        alpha=cageAlpha
    )

    ax.set_xlabel(r'$x$ [$\mathrm{\AA}$]')
    ax.set_ylabel(r'$y$ [$\mathrm{\AA}$]')
    ax.set_zlabel(r'$z$ [$\mathrm{\AA}$]')

    return fig, ax
